#define NUM_ITERATIONS 16
#define NUM_ITERATIONS_RELIEF1 11
#define NUM_ITERATIONS_RELIEF2 5

float3 size_source;

float2 parallax_occulsion( float2 uv0, float3 eyeVec, float3 lightVec, float3 normalVec, sampler2D cone_map )
{
	// POM Configurations
	int LODThreshold = 3;
	int minSamples = (int)(NUM_ITERATIONS * 0.5);
	int maxSamples = (int)(NUM_ITERATIONS * 2);
	
	float2 absoluteTexcoord = uv0 * size_source;
	float2 dxAbsolute = ddx(absoluteTexcoord);
	float2 dyAbsolute = ddy(absoluteTexcoord);
	
	// Calculate mipmap
	// mipmap = max( sqrt((Du/Dx)^2 + (Dv/Dx)^2), sqrt((Du/Dy)^2 + (Dv/Dy)^2) )
	float2 dxdySquared = dxAbsolute * dxAbsolute + dyAbsolute * dyAbsolute;
	float texcoordDelta = max(dxdySquared.x, dxdySquared.y);
	
	//float mipmap = max(0.0, 0.5 * log2(texcoordDelta));
	float mipmapLevel = max(0.0f, sqrt(texcoordDelta));
	
	float2 dx = ddx(uv0);
	float2 dy = ddy(uv0);
	
	float3 rayPos = float3(uv0, 0.0);
	if (mipmapLevel <= (float)LODThreshold)
	{
		float3 rayVec = normalize(eyeVec);
		int numSteps = (int)lerp(maxSamples, minSamples, dot(rayVec, normalize(normalVec)));
		
		rayVec.z = abs(rayVec.z);
		if (DEPTH_BIAS)
		{
			float db = 1.0 - rayVec.z;
			db *= db;
			db = 1.0 - db*db;
			rayVec.xy *= db;
		}
		rayVec.xy *= depth;
		rayVec /= rayVec.z*numSteps;
		int i;
		for(i=0; i < numSteps; i++)
		{
			float4 tex = tex2Dgrad(cone_map, rayPos.xy, dx, dy);
			if (rayPos.z < tex.w)
			{
				rayPos+=rayVec;
			}
		}
	}
	
	return rayPos.xy;
}

float2 relief(float2 uv0, float3 eyeVec, float3 lightVec, sampler2D cone_map) 
{
	float3 rayPos;
	float3 rayVec;
	
	rayPos = float3(uv0, 0.0);
	rayVec = normalize(eyeVec);
	rayVec.z = abs(rayVec.z);

	// Depth bias used by Fabio Policarpo
	if (DEPTH_BIAS)
	{
		float db = 1.0 - rayVec.z;
		db *= db;
		db = 1.0 - db*db;
		rayVec.xy *= db;
	}
	
	rayVec.xy *= depth;

	// Discretize the space
	rayVec /= rayVec.z * NUM_ITERATIONS_RELIEF1;
	
	// Linear serach
	for(int i = 0; i < NUM_ITERATIONS_RELIEF1; i++ )
	{
		float4 tex = tex2D(cone_map, rayPos.xy);
		if (rayPos.z < tex.w)
			rayPos+=rayVec;
	}
	
	// Binary seracg
	for(int i = 0; i < NUM_ITERATIONS_RELIEF2; i++)
	{
		rayVec *= 0.5;
		float4 tex = tex2D(cone_map, rayPos.xy);
		if (rayPos.z < tex.w)
			rayPos += rayVec;
		else
			rayPos -= rayVec;
	}
	
	return rayPos.xy;
}

float2 parallax(float2 uv0, float3 eyeVec, float3 lightVec, sampler2D cone_map) 
{
	// Parallax Offset Mapping
	float parallaxScale = 0.06;
	float parallaxBias = 0.015;

	float offset = tex2D(cone_map, uv0).w * parallaxScale - parallaxBias;
	float2 rayPos = uv0;
	rayPos += normalize(eyeVec) * offset;
	
	return rayPos;
}

float2 cone(float2 uv0, float3 eyeVec, sampler2D cone_map)
{
	float3 rayPos;
	float3 rayVec;
	rayPos = float3(uv0, 0.0);
	
	rayVec = normalize(eyeVec);
	rayVec.z = abs(rayVec.z);

	// DerayPosth bias used by Fabio rayPosolicarrayPoso
	if (DEPTH_BIAS)
	{
		float db = 1.0 - rayVec.z;
		db *= db;
		db = 1.0 - db*db;
		rayVec.xy *= db;
	}
	
	rayVec.xy *= depth;

	float dist = length(rayVec.xy);
	
	for( int i=0;i<NUM_ITERATIONS; i++ )
	{
		float4 tex = tex2D(cone_map, rayPos.xy);
		float height = saturate(tex.w - rayPos.z);

		float cone_ratio = tex.z*tex.z;
		
		//float stepDist = height * cone_ratio / (rayVec.z*cone_ratio + dist);
		float stepDist = height * cone_ratio / (cone_ratio + dist);
		rayPos += rayVec * stepDist;
		
		///1
		// float4 tex = tex2D(cone_map, rayPos.xy);
		// float height = max(0, tex.w - rayPos.z);
		// float stepDist = height * 0.2;
		// rayPos += rayVec * stepDist;
		
		//float stepDist = height * cone_ratio / (cone_ratio + 0.1);
		///2
		//rayPos += rayVec * height * cone_ratio / (rayVec.z*cone_ratio + dist);
		
	}
	
	return rayPos.xy;
}

float2 cone3D_fixed(float2 uv0, float3 eyeVec, sampler3D cone_map) 
{
	float3 rayPos;
	float3 rayVec;
	rayPos = float3(uv0, 0.0);
	
	rayVec = normalize(eyeVec);
	rayVec.z = abs(rayVec.z);

	// DerayPosth bias used by Fabio rayPosolicarrayPoso
	if (DEPTH_BIAS)
	{
		float db = 1.0 - rayVec.z;
		db *= db;
		db = 1.0 - db*db;
		rayVec.xy *= db;
	}
	
	rayVec.xy *= depth;
	
	for( int i=0;i<NUM_ITERATIONS;i++ )
	{
		float stepDist = tex3D(cone_map, rayPos);
		rayPos += rayVec * stepDist;
	}


	
	// float iz = max(abs(v.x),abs(v.y));// Lonesock
  
	// for( int i=0;i<NUM_ITERATIONS; i++ )
	// {
		// float4 tex = tex2D(cone_map, rayPos.xy);
		
		// float radio = tex.z*tex.z;
		// float height = tex.w - rayPos.z;
		// float dist = 1.2f * height/(iz/(radio) + 1.0);
		// rayPos += v*dist;
	// }
	return rayPos.xy;
	
}

float2 sphere(float2 uv0, float3 eyeVec, sampler3D cone_map) 
{
	float3 rayPos;
	float3 rayVec;
	rayPos = float3(uv0, 0.0f);
	
	// Texture width MUST be equal to texture height
	//****eyeVec.xy *= depth * 15.0f/256.0f * size_source.x;

	rayVec = normalize(eyeVec);
	rayVec.z = abs(rayVec.z);

	if (DEPTH_BIAS)
	{
		float db = 1.0 - rayVec.z;
		db *= db;
		db = 1.0 - db*db;
		rayVec.xy *= db;
	}
	
	rayVec.xy *=  size_source.z / size_source.x;

	for(int i = 0; i < NUM_ITERATIONS; i++)
	{
		float stepDist = tex3D(sphere_map, rayPos).x;
		rayPos += rayVec * stepDist;
	}

	return rayPos.xy;
}